/*    Clock Firmware for the ESP32 series of boards and nixie clock PN Z57xM6DV2 available at www.neonixie.com/Z57xM6DV2/
 *    Copyright © 2023 Neonixie.com, All Spectrum
 *    Permission for use and modification is provided for clocks purchased from us.
 *    Software is provided as-is, with no warranty expressed or implied.
 *    
 *    YOU SHOULD NOT PUT THIS DEVICE ON THE PUBLIC INTERNET - KEEP IT BEHIND A ROUTER/FIREWALL!
 *    
 *    Auth info
 *    - Web based firmware update:
 *            Z57XM6DV2:  http username = "z57clock"   password = "neonixie.com";
 *            IV-9-6D-RR: http username = "iv9clock"   password = "neonixie.com";
 *    - Arduino OTA firmware update, username = "neonixie.com"
 *    - mDNS hostname, usually used with Arduino OTA or local network:   
 *            Z57XM6DV2:  Z57clock-MAC_ADDRESS (Z57clock-MAC_ADDRESS.local)
 *            IV-9-6D-RR: iv9clock-MAC_ADDRESS (iv9clock-MAC_ADDRESS.local)
 *    - SSID when clock is in AP mode/SETUP mode (or when unable to connect to configured access point):
 *            Z57XM6DV2:  SSID = "z57clock"   password = "neonixie.com"
 *            IV-9-6D-RR: SSID = "iv9clock"   password = "neonixie.com"
 *            
 *    Revision History
 *    0.90  07-31-2022  First public release
 *    
 *    0.91  not final   NOTE: Version 0.91 BETA4 is considered release quality since no major bugs have been found   
 *                      Add Cathode poisoning prevention options
 *                      Add date display options
 *                      Add leading zero blanking option
 *                      Add AutoConnect OTA Web based uploading of new firmware
 *                      Add temperature and humidity compensation options
 *                      Add temperature display every X seconds and duration options, removed temperature display tickbox
 *                      Add correct temperature display code for degree F at or above 100
 *                      Moved struct tm timeinfo; to GLOBAL
 *                      Add total uptime in days and total cumulative runtime in days (stored in non-volatile flash)
 *                      Add temperature, humidity, lux, version, uptime and total runtime to main web page on device
 *                      Code cleanup, move original time code (before timezone support) to extra-code.txt
 *                      Changed bootup IP address display to full IP address (prior code only displayed last two octets)
 *                      Added code to reset permanent (flash / like eeprom) variables in case of corruption or incorrect values cause a crash that we cannot recover from
 *                        This is done by grounding pin 15 on the ESP while the board is reset/powered up. The display will flash all 0's three times to indicate 
 *                        that flash variables have been reset to default values. TOTAL_RUNTIME is preserved.
 *                      Added watchdog timer of 300 seconds. Mainly to force the clock to attempt to reconnect to wifi when in soft AP mode. Should not get triggered in any other mode.
 *                      Added config.autoReconnect = true; to make wifi attempt a reconnect to known routers. Without this if the wifi router went offline, the clock would not reconnect.
 *                      BETA4, Removed a return from initTime, was causing timezone not to set sometimes at initial bootup
 *                      BETA4, Removed free memory display on main http page
 *                      BETA4, Changed light sensor default to off in eeprom
 *                      BETA4, Added default brightness variable in setup(), without this if light sensor was selected and none was present display would be blank
 *                      BETA4, Change default sleep time to display IP address from 2s to 4s
 *                      BETA4, Removed extra getLocalTime function and a few lines around it in cathodepp() function
 *                      BETA4, Modified ArduinoOTA.setHostname to shorten local hostname, was full MAC now Z57clock-[last-two-octets-of-MAC]
 *                      
 *    0.92 BETA1 UNIV   FIRST UNIVERSAL CODEBASE FOR IV-9-6D-RR and Z57XM6DV2. Sourced from Z57XM6DV2 code v0.91 BETA4
 *                      Arduino IDE 1.8.19, ESP32 board manager v2.0.6, AutoConnect v1.4.0, Adafruit_BMP280_Library v2.6.6, Adafruit_AHTX0 v2.0.1
 *                      MAIN CHANGES:
 *                      Added preprocessor defines for clock type and autodetection of board type used
 *                      Several serial_debugxxx preprocessor defines added for Serial.Printxxx statements to remove serial output when not needed
 *                      Driver mapping include file name changed to driver-mapping-definitions.h
 *                      Added full alphabet support to driver-mapping-definitions.h for IV-9 clock
 *                      HTML config page text moved to html-definitions.h
 *                      Added esp_partition.h and related code to display partition information, mainly sizes, linked at main http page
 *                      Added code to read BMP280, added read options on configuration page and display results on main web page
 *                      Added code to enable/disable colon indicators to main and config web pages
 *                      Added code to handle "Tubes on bottom of PCB" to main and config web pages for IV-9 clock
 *                      Added code for "Alternate Display" option. Options are standard (clock), altimeter, RSSI, Lux and Barometer (HG)
 *                      Moved decimal point code outside of write_display function and related changes
 *                      PWM dimming set to 1,600 hz
 *                      Added code to set default brightness in settings to 165 for IV-9-6D-RR and 255 for Z57XM6DV2
 *                      Removed delay(20) in main loop
 *                      Changes made to various hostname settings to fix issues caused by upgrading the ESP32 board manager
 *                        When upgrading the ESP32 board manager from 1.0.6 to 2.0.5 hostnames (softAP hostname, wifi/dhcp hostname and MDNS) were initially not setting correctly, they would be ESP32xxxx defaults
 *                        now working as before after several code modifications
 *                      
 *                      
 *                      
 *                      
 */
// WIFI CONFIG / CONNECTION / WEB SERVER BASED on https://hieromon.github.io/AutoConnect/
// NTP BASED on https://RandomNerdTutorials.com/esp8266-nodemcu-date-time-ntp-client-server-arduino/
// EEPROM / "PREFERENCES" https://randomnerdtutorials.com/esp32-save-data-permanently-preferences/

//  THIS IS WHAT IS NORMALLY USED FOR DEV using original esp32-devkitc-ve BOARD, Also works with ESP32-DevKitC-32D
// Board Node32s    or    DOIT ESP32 DEVKIT V1  seems to work too! // upload speed 921600 // flash freq 80mhz // part scheme default
// 01/13/2023 BOARD UPDATE esp32-devkitc-ve works well as "ESP32 Wrover Kit (all versions)" and has more board options to choose from

const char VERSION[] = "0.92 BETA1 UNIVERSAL";
const char compile_date[] = __DATE__ " " __TIME__;

// TYPE OF CLOCK, ONLY UNCOMMENT ONE
#define Z57XM6DV2
//#define IV_9_6D_RR

#ifdef IV_9_6D_RR 
  #define CLOCKMODEL "IV-9-6D-RR"
  #define CLOCKTYPE "Numitron"
  #define WRITE_DISPLAY(a,b,c,d,e,f,g,h,i,j,k,l) WRITE_DISPLAY_IV9(a,b,c,d,e,f,g,h,i,j,k,l)
  #define DIGIT_NEGATIVE 36
  #define DIGIT_BLANK 37
  #ifdef ARDUINO_ADAFRUIT_QTPY_ESP32S3_NOPSRAM  // Works but not extensively tested, QIO 80mhz, Flash Size 8MB, Partition Default (3MB APP/1.5MB SPIFFS), Arduino and Events on Core 1
    #define CLOCK_PIN 18  // A0
    #define RCK_PIN 17    // A1
    #define OE_PIN 9      // A2
    #define DATA_PIN 8    // A3 
    #define EEPROM_RESET_PIN 16
  #endif
  #ifdef ARDUINO_ADAFRUIT_QTPY_ESP32C3          // Works but not extensively tested, QIO 80mhz, Flash Size 4MB, Minimal SPIFFS (1.9MB APP OTA / 190KB SPIFFS)
    #define CLOCK_PIN 4   // A0
    #define RCK_PIN 3     // A1
    #define OE_PIN 1      // A2
    #define DATA_PIN 0    // A3
    #define EEPROM_RESET_PIN 20
  #endif
  #ifdef ARDUINO_XIAO_ESP32C3 // Flash size 4MB, Minimal SPIFFS (1.9MB APP OTA / 190KB SPIFFS), Flash FREQ 80mhz, Flash Mode QIO
    #define CLOCK_PIN 2   // A0
    #define RCK_PIN 3     // A1
    #define OE_PIN 4      // A2
    #define DATA_PIN 5    // A3
    #define EEPROM_RESET_PIN 20
  #endif  
  #ifdef ARDUINO_ESP32_WROVER_KIT // THIS IS "ESP32 Wrover Kit (all versions)" in the Arduino IDE, flash freq 80mhz, QIO, 8mb flash size, 8M with spiffs (3MB APP/1.5MB SPIFFS), PSRAM enabled
    #define CLOCK_PIN 19
    #define RCK_PIN 18
    #define OE_PIN 2
    #define DATA_PIN 4
    #define EEPROM_RESET_PIN 15
  #endif
//  #ifdef ARDUINO_ESP32_DEV    // THIS IS "ESP32 Wrover Module" in the Arduino IDE, TEST ONLY, USE INSTEAD "ESP32 Wrover Kit (all versions)"
//    #define OE_PIN 2
//    #define DATA_PIN 4
//    #define CLOCK_PIN 19
//    #define RCK_PIN 18
//  #endif
#endif

#ifdef Z57XM6DV2    // Z57XM6DV2 CURRENTLY ONLY USES ONE BOARD, SO NO NEED TO CHECK BOARD TYPE 
  #define CLOCKMODEL "Z57XM6DV2"
  #define CLOCKTYPE "Nixie"
  #define WRITE_DISPLAY(a,b,c,d,e,f,g,h,i,j,k,l) WRITE_DISPLAY_Z57(a,b,c,d,e,f,g,h,i,j,k,l)
  #define DIGIT_NEGATIVE 10 // WE DON'T HAVE A NEGATIVE ON A NIXIE, NOT THIS ONE AT LEAST! SO WE JUST SET IT TO BLANK
  #define DIGIT_BLANK 10
  #define OE_PIN 19
  #define CLOCK_PIN 18
  #define DATA_PIN 4
  #define EEPROM_RESET_PIN 15
#endif

// 1068881 with serial debug on
// 1067129 with serial debug off, size comparison, a difference of 1,752 bytes, tested with pre-release version
#define SERIAL_DEBUG 0

#if SERIAL_DEBUG == 1
  #define serial_debug(x) Serial.print(x)
  #define serial_debug_x2(x,y) Serial.print(x,y)
  #define serial_debugln(x) Serial.println(x)
  #define serial_debug_printf(x,y) Serial.printf(x,y)
#else
  #define serial_debug(x)
  #define serial_debug_x2(x,y)
  #define serial_debugln(x)
  #define serial_debug_printf(x,y)
#endif

#include <WiFi.h>
#include <NTPClient.h>
#include <WiFiUdp.h>
#include <Wire.h>
#include <BH1750.h>
#include <Preferences.h>    // FOR "EEPROM" SAVING, IT'S CALLED "Preferences" in Arduino
Preferences preferences;

#include "driver-mapping-definitions.h"    // Mapping for HV5122 shift registers.. SEC_ONE_MAP, SEC_TEN_MAP, etc... and IV-9 driver ICs
#include "html-definitions.h"    // Web page code

// OTA STUFF
#include <ArduinoOTA.h>
#include <ESPmDNS.h>

// FOR ESP PARTITION READING
#include "esp_partition.h"

// Watchdog timer
#include <esp_task_wdt.h>

// AHT20 CODE BASED ON https://github.com/adafruit/Adafruit_AHTX0
// INSTALL FROM LIBRARY MANAGER AND SEARCH FOR AHT20, install dependencies as well
#include <Adafruit_AHTX0.h>
Adafruit_AHTX0 aht;

// BMP280 CODE https://github.com/adafruit/Adafruit_BMP280_Library
// INSTALL IN LIBRARY MANAGER AND SEARCH FOR "Adafruit BMP280"
#define BMP_ENABLE
#ifdef BMP_ENABLE
  #include <Adafruit_BMP280.h>
  Adafruit_BMP280 bmp; // I2C
#endif
float bmp_PA;
float bmp_HG;
char bmp_HG_STRING_BUFFER[8];
float bmp_METERS;
float bmp_FEET;
float bmp_TEMPERATURE_C;

struct tm timeinfo;
char TIME_STRING[150];

float HUMIDITY;
float TEMPERATURE_C;
float TEMPERATURE_F;

float lux;

char tmp_hostname[30];

int SETTING_BRIGHTNESS;
int SETTING_LUXMETER;
int SETTING_12H;
int SETTING_LZBLANK;
int SETTING_DATEFORMAT;
int SETTING_DATEDISP_SEC;
int SETTING_DATEDISP_DUR;
String SETTING_TIMEZONE;
//int SETTING_DISPLAYTEMP;
int SETTING_TEMP_IN_C;
int SETTING_TEMPDISP_SEC;
int SETTING_TEMPDISP_DUR;
int SETTING_TEMP_COMP;
int SETTING_HUMIDITY_COMP;
int SETTING_CATHODEPP;
int SETTING_CATHODEPPSTART;
int SETTING_CATHODEPPDUR;
int SETTING_CATHODEPPREV;
int SETTING_CATHODEPPREVD;

int SETTING_TUBEONBOTTOM;

int SETTING_COLON;

int SETTING_ALTDISPLAY;

int SETTING_BMP;

int SETTING_TEMPHUM;

int BRIGHTNESS;

float TOTAL_RUNTIME;   // TOTAL TIME CLOCK HAS BEEN POWERED UP IN DAYS
int TOTAL_RUNTIME_SECONDS_COUNTER = 0;
float UPTIME = 0;       // UPTIME IN DAYS

BH1750 lightMeter;

// FOR TIMEZONE???  // FROM     https://randomnerdtutorials.com/esp32-ntp-timezones-daylight-saving/
#include "time.h"

// C:\Users\$YOUR-USERNAME\Documents\Arduino\libraries\AutoConnect\src\AutoConnectDefs.h
// Uncomment the following AC_DEBUG to enable debug output. IN THE ABOVE FILE!!!!!!!!
// #define AC_DEBUG ////////////// IN THE ABOVE FILE!!!!!!!!
#include <AutoConnect.h>  // CHECK http://IP/_ac  when connected!
#include <WebServer.h>
WebServer Server;
AutoConnect Portal(Server);
// ORIGINAL THAT DOES NOT WORK ANYMORE ON NEWER VERSION 12-25-2022, SEE PSK STUFF BELOW
// BUT MUST HAVE IT TO WORK?? something with config variable?
//AutoConnectConfig config("z57clock", "neonixie.com");    // This is the default SSID and PASSWORD used when the device can't connect and goes to "AP" mode
AutoConnectConfig config;

void rootPage() {
  char content[1200]; // MAKE SURE content[] is big enough!
  
  // Changed sprintf below to snprintf so we don't crash if the string overflows
  snprintf(content, sizeof(content), "Hello, This is a %s clock. Model %s from www.neonixie.com!<br>"\
  "To configure settings click <a href=/clockconfig>here</a><br>"\
  "Online documentation is available <a href=http://www.neonixie.com/%s/?s=fromaclock>Here</a><br>"\
  "<br>"\
  "Temp C: %.2f Temp F: %.2f Humidity: %.0f<br>"\
  "BMP280 sensor PA: %.2f HG: %.2f Meters: %.2f Feet: %.1f Temperature_C %.1f<br>"\
  "Lux: %.0f<br>"\
  "<br>"\
  "IP = %s<br>"\
  "mDNS Hostname = %s.local<br>"\
  "ARDUINO_BOARD = %s<br>"\
  "Version %s compile time: %s <a href=/partitions.html>Partitions</a><br>"\
  "Uptime Seconds: %lld  Uptime Days: %.2f  Total Runtime Days: %.2f<br>"\
  "%s<br>"\
  , CLOCKTYPE, CLOCKMODEL, CLOCKMODEL, TEMPERATURE_C, TEMPERATURE_F, HUMIDITY, bmp_PA, bmp_HG, bmp_METERS, bmp_FEET, bmp_TEMPERATURE_C, lux, WiFi.localIP().toString().c_str(), tmp_hostname, ARDUINO_BOARD, VERSION, compile_date, esp_timer_get_time()/1000000, UPTIME, TOTAL_RUNTIME, TIME_STRING);
  
  // removed in 0.91 BETA4... "Free memory: %d" ...  ESP.getFreeHeap() 
  Server.send(200, "text/html", content);
}

// Define NTP Client to get time
WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP, "pool.ntp.org");

int sec_ones, sec_tens, min_ones, min_tens, hour_ones, hour_tens;
byte sec_ones_dp, sec_tens_dp, min_ones_dp, min_tens_dp, hour_ones_dp, hour_tens_dp;

int previous_tm_sec;  // USE THIS TO SEE WHEN A SECOND TICKS OVER

int PREVIOUS_TEMPERATURE_READ_SECONDS;

//AutoConnect portal;
AutoConnectAux  page_add;
//AutoConnectAux  page_results;

int INITIALIZE_WEB_VARIABLES=0;

String onclockconfig(AutoConnectAux& aux, PageArgument& args) {
  // IMPORTANT NOTE!!! IF THE SETTING VARIABLES BELOW ARE NOT ON THE WEBPAGE THE PROGRAM WILL CRASH WHEN YOU TRY TO LOAD THE PAGE!!!
  if ( INITIALIZE_WEB_VARIABLES == 0 ) {
    aux["SETTING_BRIGHTNESS"].as<AutoConnectInput>().value = SETTING_BRIGHTNESS;
    //aux["SETTING_LUXMETER"].as<AutoConnectCheckbox>().checked = SETTING_LUXMETER;
    aux["SETTING_LUXMETER"].as<AutoConnectSelect>().selected = SETTING_LUXMETER;
    
    //aux["SETTING_DISPLAYTEMP"].as<AutoConnectCheckbox>().checked = SETTING_DISPLAYTEMP;
    aux["SETTING_TEMP_IN_C"].as<AutoConnectCheckbox>().checked = SETTING_TEMP_IN_C;
    aux["SETTING_TEMPDISP_SEC"].as<AutoConnectInput>().value = SETTING_TEMPDISP_SEC;
    aux["SETTING_TEMPDISP_DUR"].as<AutoConnectInput>().value = SETTING_TEMPDISP_DUR; 
    aux["SETTING_TEMP_COMP"].as<AutoConnectInput>().value = SETTING_TEMP_COMP;
    aux["SETTING_HUMIDITY_COMP"].as<AutoConnectInput>().value = SETTING_HUMIDITY_COMP;
    
    aux["SETTING_12H"].as<AutoConnectCheckbox>().checked = SETTING_12H;
    aux["SETTING_LZBLANK"].as<AutoConnectCheckbox>().checked = SETTING_LZBLANK;

    aux["SETTING_DATEFORMAT"].as<AutoConnectSelect>().selected = SETTING_DATEFORMAT;
    aux["SETTING_DATEDISP_SEC"].as<AutoConnectInput>().value = SETTING_DATEDISP_SEC;
    aux["SETTING_DATEDISP_DUR"].as<AutoConnectInput>().value = SETTING_DATEDISP_DUR;

    #ifdef Z57XM6DV2
      aux["SETTING_CATHODEPP"].as<AutoConnectSelect>().selected = SETTING_CATHODEPP;
      aux["SETTING_CATHODEPPSTART"].as<AutoConnectInput>().value = SETTING_CATHODEPPSTART;
      aux["SETTING_CATHODEPPDUR"].as<AutoConnectInput>().value = SETTING_CATHODEPPDUR;
      aux["SETTING_CATHODEPPREV"].as<AutoConnectInput>().value = SETTING_CATHODEPPREV;
      aux["SETTING_CATHODEPPREVD"].as<AutoConnectInput>().value = SETTING_CATHODEPPREVD;
    #endif
    
    aux["SETTING_TIMEZONE"].as<AutoConnectInput>().value = SETTING_TIMEZONE;

    #ifdef IV_9_6D_RR
      aux["SETTING_TUBEONBOTTOM"].as<AutoConnectSelect>().selected = SETTING_TUBEONBOTTOM;
    #endif

    aux["SETTING_COLON"].as<AutoConnectSelect>().selected = SETTING_COLON;

    aux["SETTING_ALTDISPLAY"].as<AutoConnectSelect>().selected = SETTING_ALTDISPLAY;
    
    aux["SETTING_BMP"].as<AutoConnectSelect>().selected = SETTING_BMP;

    aux["SETTING_TEMPHUM"].as<AutoConnectSelect>().selected = SETTING_TEMPHUM;
    
    aux["somemoretextbla1"].as<AutoConnectText>().value = "Version " + String(VERSION) + " compile time: " + String(compile_date) + "<br>";   
    INITIALIZE_WEB_VARIABLES = 1;
  } else {

    preferences.begin("clock-settings", false);
    
    SETTING_BRIGHTNESS = aux["SETTING_BRIGHTNESS"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("BRIGHTNESS", SETTING_BRIGHTNESS);
    
    //SETTING_LUXMETER = aux["SETTING_LUXMETER"].as<AutoConnectCheckbox>().checked;
    SETTING_LUXMETER = aux["SETTING_LUXMETER"].as<AutoConnectSelect>().selected;
    preferences.putInt("LUXMETER", SETTING_LUXMETER);
    
    //SETTING_DISPLAYTEMP = aux["SETTING_DISPLAYTEMP"].as<AutoConnectCheckbox>().checked;
    //preferences.putInt("DISPLAYTEMP", SETTING_DISPLAYTEMP);
    SETTING_TEMP_IN_C = aux["SETTING_TEMP_IN_C"].as<AutoConnectCheckbox>().checked;
    preferences.putInt("TEMP_IN_C", SETTING_TEMP_IN_C); 
    SETTING_TEMPDISP_SEC = aux["SETTING_TEMPDISP_SEC"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("TEMPDISP_SEC", SETTING_TEMPDISP_SEC);
    SETTING_TEMPDISP_DUR = aux["SETTING_TEMPDISP_DUR"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("TEMPDISP_DUR", SETTING_TEMPDISP_DUR);  
    
    SETTING_TEMP_COMP = aux["SETTING_TEMP_COMP"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("TEMP_COMP", SETTING_TEMP_COMP);
    SETTING_HUMIDITY_COMP = aux["SETTING_HUMIDITY_COMP"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("HUMIDITY_COMP", SETTING_HUMIDITY_COMP);

    SETTING_12H = aux["SETTING_12H"].as<AutoConnectCheckbox>().checked;
    preferences.putInt("12H", SETTING_12H);

    SETTING_LZBLANK = aux["SETTING_LZBLANK"].as<AutoConnectCheckbox>().checked;
    preferences.putInt("LZBLANK", SETTING_LZBLANK);

    SETTING_DATEFORMAT = aux["SETTING_DATEFORMAT"].as<AutoConnectSelect>().selected;
    preferences.putInt("DATEFORMAT", SETTING_DATEFORMAT);
    SETTING_DATEDISP_SEC = aux["SETTING_DATEDISP_SEC"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("DATEDISP_SEC", SETTING_DATEDISP_SEC);
    SETTING_DATEDISP_DUR = aux["SETTING_DATEDISP_DUR"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("DATEDISP_DUR", SETTING_DATEDISP_DUR);    

    #ifdef Z57XM6DV2
      SETTING_CATHODEPP = aux["SETTING_CATHODEPP"].as<AutoConnectSelect>().selected;
      preferences.putInt("CATHODEPP", SETTING_CATHODEPP);
      SETTING_CATHODEPPSTART = aux["SETTING_CATHODEPPSTART"].as<AutoConnectInput>().value.toInt();
      preferences.putInt("CATHODEPPSTART", SETTING_CATHODEPPSTART);
      SETTING_CATHODEPPDUR = aux["SETTING_CATHODEPPDUR"].as<AutoConnectInput>().value.toInt();
      preferences.putInt("CATHODEPPDUR", SETTING_CATHODEPPDUR);
      SETTING_CATHODEPPREV = aux["SETTING_CATHODEPPREV"].as<AutoConnectInput>().value.toInt();
      preferences.putInt("CATHODEPPREV", SETTING_CATHODEPPREV);
      SETTING_CATHODEPPREVD = aux["SETTING_CATHODEPPREVD"].as<AutoConnectInput>().value.toInt();
      preferences.putInt("CATHODEPPREVD", SETTING_CATHODEPPREVD);
    #endif

    SETTING_TIMEZONE = aux["SETTING_TIMEZONE"].as<AutoConnectInput>().value;
    preferences.putString("TIMEZONE", SETTING_TIMEZONE);

    #ifdef IV_9_6D_RR 
      SETTING_TUBEONBOTTOM = aux["SETTING_TUBEONBOTTOM"].as<AutoConnectSelect>().selected;
      preferences.putInt("TUBEONBOTTOM", SETTING_TUBEONBOTTOM);
    #endif

    SETTING_COLON = aux["SETTING_COLON"].as<AutoConnectSelect>().selected;
    preferences.putInt("COLON", SETTING_COLON);

    SETTING_BMP = aux["SETTING_BMP"].as<AutoConnectSelect>().selected;
    preferences.putInt("BMP", SETTING_BMP);

    SETTING_TEMPHUM = aux["SETTING_TEMPHUM"].as<AutoConnectSelect>().selected;
    preferences.putInt("TEMPHUM", SETTING_TEMPHUM);

    SETTING_ALTDISPLAY = aux["SETTING_ALTDISPLAY"].as<AutoConnectSelect>().selected;
    preferences.putInt("ALTDISPLAY", SETTING_ALTDISPLAY);
    
    setTimezone(SETTING_TIMEZONE);
    
    preferences.end();
  }

  return String();
}

String onResults(AutoConnectAux& aux, PageArgument& args) {
  int valueA = args.arg("valueA").toInt();
  int valueB = args.arg("valueB").toInt();

  ledcWrite(0, valueA);

  aux["results"].as<AutoConnectText>().value = String(valueA) + " + " + String(valueB) + " = " + String(valueA + valueB);
  return String();
}
// END ADD PAGE TEST

void WRITE_BITS(int x) {
  #ifdef IV_9_6D_RR
  int var = 7;
  #else
  int var = 9;
  #endif
  
  while (var >= 0) {
    digitalWrite(CLOCK_PIN, 0);
    //digitalWrite(DATA_PIN, (x >> var) & 0x01);    // REGULAR
    digitalWrite(DATA_PIN, !((x >> var) & 0x01));      // INVERTED
    digitalWrite(CLOCK_PIN, 1);
    var--;
  }
}

void WRITE_ONELOWBIT() {
    digitalWrite(CLOCK_PIN, 0);
    digitalWrite(DATA_PIN, 1);
    digitalWrite(CLOCK_PIN, 1);
}

void WRITE_ONEHIGHBIT() {
    digitalWrite(CLOCK_PIN, 0);
    digitalWrite(DATA_PIN, 0);
    digitalWrite(CLOCK_PIN, 1);
}

#ifdef IV_9_6D_RR
byte BITS_FOR_THE_UPSIDE_DOWN(byte TOP) // THIS IS FOR IV-9 clock where tubes are mounted on the bottom of the PCB
{
  byte BOTTOM = 0;

  bitWrite(BOTTOM, 0, bitRead(TOP,0));
  bitWrite(BOTTOM, 1, bitRead(TOP,7));
  bitWrite(BOTTOM, 2, bitRead(TOP,6));
  bitWrite(BOTTOM, 3, bitRead(TOP,5));
  bitWrite(BOTTOM, 4, bitRead(TOP,4));
  bitWrite(BOTTOM, 5, bitRead(TOP,3));
  bitWrite(BOTTOM, 6, bitRead(TOP,2));
  bitWrite(BOTTOM, 7, bitRead(TOP,1));

  return BOTTOM;
}
#endif

#ifdef IV_9_6D_RR
void WRITE_DISPLAY_IV9(int HT, int HO, int MT, int MO, int ST, int SO, byte HT_DP, byte HO_DP, byte MT_DP, byte MO_DP, byte ST_DP, byte SO_DP)
{
  if ( SETTING_TUBEONBOTTOM == 2 ) {  // THIS IS FOR IV-9 clock where tubes are mounted on the bottom of the PCB
    WRITE_BITS((SO_DP ^ BITS_FOR_THE_UPSIDE_DOWN(IV9_DIGIT_MAP[SO])));
    WRITE_BITS((ST_DP ^ BITS_FOR_THE_UPSIDE_DOWN(IV9_DIGIT_MAP[ST])));
    WRITE_BITS((MO_DP ^ BITS_FOR_THE_UPSIDE_DOWN(IV9_DIGIT_MAP[MO])));
    WRITE_BITS((MT_DP ^ BITS_FOR_THE_UPSIDE_DOWN(IV9_DIGIT_MAP[MT])));
    WRITE_BITS((HO_DP ^ BITS_FOR_THE_UPSIDE_DOWN(IV9_DIGIT_MAP[HO])));
    WRITE_BITS((HT_DP ^ BITS_FOR_THE_UPSIDE_DOWN(IV9_DIGIT_MAP[HT])));    
  } else {
    WRITE_BITS((HT_DP ^ IV9_DIGIT_MAP[HT]));
    WRITE_BITS((HO_DP ^ IV9_DIGIT_MAP[HO]));
    WRITE_BITS((MT_DP ^ IV9_DIGIT_MAP[MT]));
    WRITE_BITS((MO_DP ^ IV9_DIGIT_MAP[MO]));
    WRITE_BITS((ST_DP ^ IV9_DIGIT_MAP[ST]));
    WRITE_BITS((SO_DP ^ IV9_DIGIT_MAP[SO]));
  }
     
  digitalWrite(RCK_PIN, 1);
  digitalWrite(RCK_PIN, 0);
}
#endif

#ifdef Z57XM6DV2
void WRITE_DISPLAY_Z57(int HT, int HO, int MT, int MO, int ST, int SO, byte HT_DP, byte HO_DP, byte MT_DP, byte MO_DP, byte ST_DP, byte SO_DP)
{
    WRITE_BITS(MIN_TEN_MAP[MT]);

    if ( HO_DP == 1 ) { // DECIMAL POINT ON
      WRITE_ONEHIGHBIT();    // NEON BULB
    } else {
      WRITE_ONELOWBIT();    // NEON BULB
    }
    
    WRITE_BITS(HOUR_ONE_MAP[HO]);
    WRITE_BITS(HOUR_TEN_MAP[HT]);
    
    WRITE_ONELOWBIT();    // PIN 35 U4 /  NO CONNECTION

    WRITE_BITS(SEC_ONE_MAP[SO]);
    WRITE_BITS(SEC_TEN_MAP[ST]);
    
    if ( MO_DP == 1 ) { // DECIMAL POINT ON
      WRITE_ONEHIGHBIT();    // NEON BULB
    } else {
      WRITE_ONELOWBIT();    // NEON BULB
    }
    
    WRITE_BITS(MIN_ONE_MAP[MO]);
    
    WRITE_ONELOWBIT();    // PIN 35 NO CONNECTION  

    digitalWrite(CLOCK_PIN, 0);   // WRITE CLOCK PIN TO LOW TO KEEP LOGIC LEVEL SHIFTER (RESISTOR AND FET) TEMPERATURES LOWER
    digitalWrite(DATA_PIN, 0);    // WRITE DATA PIN TO LOW TO KEEP LOGIC LEVEL SHIFTER (RESISTOR AND FET) TEMPERATURES LOWER
}
#endif

void demo_digits(){
    int var = 0;
    #ifdef IV_9_6D_RR
      while (var <= 9) { // SET TO 37 TO DISPLAY THE WHOLE ALPHABET!
    #else 
      while (var <= 9) {
    #endif
      WRITE_DISPLAY(var, var, var, var, var, var, 0, 0, 0, 0, 0, 0);
      var++;
      delay(500);      
    }
  //WRITE_DISPLAY(9,9,9,9,9,9, 0, 0, 0, 0, 0, 0);   
}

void setTimezone(String timezone){
  serial_debug_printf(" Setting Timezone to %s\n",timezone.c_str());
  setenv("TZ",timezone.c_str(),1);  //  Now adjust the TZ.  Clock settings are adjusted to show the new local time
  tzset();
}

void initTime(String timezone){
  //struct tm timeinfo; // Moved to global

  serial_debugln("Setting up time");
  configTime(0, 0, "pool.ntp.org");    // First connect to NTP server, with 0 TZ offset
  if(!getLocalTime(&timeinfo)){
    serial_debugln(" Failed to obtain time");
    //return; // REMOVED AS OF 0.91 BETA4, this was at times causing timezone to not set on initial bootup
  } else {
    serial_debugln(" Got the time from NTP");
  }
  setTimezone(timezone);
}

void printLocalTime(){
  //struct tm timeinfo; // Moved to global
  if(!getLocalTime(&timeinfo)){
    serial_debugln("Failed to obtain time 1");
    return;
  }
  serial_debug_x2(&timeinfo, "%A, %B %d %Y %H:%M:%S zone %Z %z ");
}

void setTime(int yr, int month, int mday, int hr, int minute, int sec, int isDst){
  struct tm tm;

  tm.tm_year = yr - 1900;   // Set date
  tm.tm_mon = month-1;
  tm.tm_mday = mday;
  tm.tm_hour = hr;      // Set time
  tm.tm_min = minute;
  tm.tm_sec = sec;
  tm.tm_isdst = isDst;  // 1 or 0
  time_t t = mktime(&tm);
  serial_debug_printf("Setting time: %s", asctime(&tm));
  struct timeval now = { .tv_sec = t };
  settimeofday(&now, NULL);
}

#ifdef Z57XM6DV2
void cathodepp (void)
{
  if ( (timeinfo.tm_hour >= SETTING_CATHODEPPSTART && timeinfo.tm_hour < (SETTING_CATHODEPPSTART+SETTING_CATHODEPPDUR)) || timeinfo.tm_hour < ((SETTING_CATHODEPPSTART+SETTING_CATHODEPPDUR)-24) ) {
    if ( SETTING_CATHODEPP == 3 ) {
       //ledcWrite(0, 0);     // SET TO FULL BRIGHTNESS
       BRIGHTNESS=255;
    }
    if ( SETTING_CATHODEPP == 4 ) {
       //ledcWrite(0, 128);   // SET TO HALF BRIGHTNESS
       BRIGHTNESS=128;
    }

    //if ( timeinfo.tm_sec == 0 || timeinfo.tm_sec == 10 || timeinfo.tm_sec == 20 || timeinfo.tm_sec == 30 || timeinfo.tm_sec == 40 || timeinfo.tm_sec == 50 ) { // OLD REMOVED 0.91 BETA4
    if ( timeinfo.tm_sec%SETTING_CATHODEPPREV < SETTING_CATHODEPPREVD ) {   // NEW 0.91 BETA4 Check if we need to revert to a regular display every SETTING_CATHODEPPREV (SECONDS) for SETTING_CATHODEPPREVD (DURATION)
      //juggleoutput(0);
    } else {
      sec_ones=timeinfo.tm_min/6;
      sec_tens=timeinfo.tm_min/6;
      min_ones=timeinfo.tm_min/6;
      min_tens=timeinfo.tm_min/6;
      hour_ones=timeinfo.tm_min/6;
      hour_tens=timeinfo.tm_min/6;
    }
              
  }
}
#endif

#ifdef BMP_ENABLE
void read_bmp() {
  if ( SETTING_BMP == 2 || (SETTING_BMP == 3 && timeinfo.tm_sec%10 == 0) ) {  // 2 is read every second (function should be in 'every second' loop, and 3 is every 10 seconds
    bmp.begin(0x77);
    // READ BMP280 PRESSURE/ALTITUDE/TEMPERATURE
    //Serial.print(F("Temperature = ")); Serial.print(bmp.readTemperature()); Serial.println(" *C");
    //Serial.print(F("Pressure = ")); Serial.print(bmp.readPressure()); Serial.println(" Pa");
    //Serial.print(F("Approx altitude = ")); Serial.print(bmp.readAltitude(1013.25)); /* Adjusted to local forecast! */ Serial.println(" m");
    bmp_PA = bmp.readPressure();
    bmp_METERS = bmp.readAltitude(1013.25); // STANDARD
    bmp_TEMPERATURE_C = bmp.readTemperature();
    bmp_HG = bmp_PA * 0.00029530;
    bmp_FEET = bmp_METERS * 3.2808;
  }
}
#endif

void display_temperature() {
  
  if ( SETTING_TEMPDISP_SEC != 0 ) {
    if ( SETTING_TEMPHUM == 2 ) { // READY EVERY SECOND
      //aht.begin();  // THIS IS MOVED OUT OF THE REGULAR LOOP
      sensors_event_t aht_humidity, aht_temp;
      aht.getEvent(&aht_humidity, &aht_temp);// populate temp and humidity objects with fresh data
      HUMIDITY=aht_humidity.relative_humidity + SETTING_HUMIDITY_COMP;
      TEMPERATURE_C=aht_temp.temperature + (SETTING_TEMP_COMP/float(10));
      TEMPERATURE_F=(TEMPERATURE_C * 9/5) + 32;
    } else if ( timeinfo.tm_sec%10 == 0 ) {    // READ TEMP AND HUMIDITY EVERY 10 SECONDS
      //aht.begin();  // THIS IS MOVED OUT OF THE REGULAR LOOP
      sensors_event_t aht_humidity, aht_temp;
      aht.getEvent(&aht_humidity, &aht_temp);// populate temp and humidity objects with fresh data
      HUMIDITY=aht_humidity.relative_humidity + SETTING_HUMIDITY_COMP;
      TEMPERATURE_C=aht_temp.temperature + (SETTING_TEMP_COMP/float(10));
      TEMPERATURE_F=(TEMPERATURE_C * 9/5) + 32;
    }
    //if ( (timeinfo.tm_sec%SETTING_TEMPDISP_SEC <= SETTING_TEMPDISP_DUR) || ((timeinfo.tm_sec%SETTING_TEMPDISP_SEC)-1) == 0 ) {
    if ( timeinfo.tm_sec%SETTING_TEMPDISP_SEC < SETTING_TEMPDISP_DUR ) {
      if ( SETTING_TEMP_IN_C == 1 ) {
          hour_tens=int(TEMPERATURE_C) / 10;
          hour_ones=int(TEMPERATURE_C) % 10;;
          min_tens=int(TEMPERATURE_C * 10) %10;
          min_ones=DIGIT_BLANK;      
          sec_tens=int(HUMIDITY) / 10;
          sec_ones=int(HUMIDITY) % 10;
          hour_tens_dp = 0; hour_ones_dp = 1; min_tens_dp = 0; min_ones_dp = 0; sec_tens_dp = 0; sec_ones_dp = 0;
      } else {
        if ( TEMPERATURE_F >= 100 ) {          // 100+F display as xxx and round
          hour_tens=int(round(TEMPERATURE_F)) %1000/100;
          hour_ones=int(round(TEMPERATURE_F)) %100/10;
          min_tens=int(round(TEMPERATURE_F)) % 10;
          min_ones=DIGIT_BLANK;
          sec_tens=int(HUMIDITY) / 10;
          sec_ones=int(HUMIDITY) % 10;
          hour_tens_dp = 0; hour_ones_dp = 0; min_tens_dp = 0; min_ones_dp = 0; sec_tens_dp = 0; sec_ones_dp = 0;
        } else {                              // LESS THEN 100F display as xx.x  
          hour_tens=int(TEMPERATURE_F) / 10;
          hour_ones=int(TEMPERATURE_F) % 10;;
          min_tens=int(TEMPERATURE_F * 10) %10;
          min_ones=DIGIT_BLANK;      
          sec_tens=int(HUMIDITY) / 10;
          sec_ones=int(HUMIDITY) % 10;
          hour_tens_dp = 0; hour_ones_dp = 1; min_tens_dp = 0; min_ones_dp = 0; sec_tens_dp = 0; sec_ones_dp = 0;
        }
      }
    }
    //Serial.print("Temp: "); Serial.print(TEMPERATURE_C); Serial.print(" C "); Serial.print(TEMPERATURE_F); Serial.print(" F "); Serial.print("Humidity: "); Serial.print(HUMIDITY); Serial.print("% rH ");
    serial_debug("Temp: "); serial_debug(TEMPERATURE_C); serial_debug(" C "); serial_debug(TEMPERATURE_F); serial_debug(" F "); serial_debug("Humidity: "); serial_debug(HUMIDITY); serial_debug("% rH ");
  } else {  // TEMPERATURE DISPLAY IS NOT ON, SO WE CHECK IF WE NEED TO READ THE SENSOR ANYWAY (FOR WEB DISPLAY PURPOSES ETC)
    if ( SETTING_TEMPHUM == 2 ) { // READY EVERY SECOND
      sensors_event_t aht_humidity, aht_temp;
      aht.getEvent(&aht_humidity, &aht_temp);// populate temp and humidity objects with fresh data
      HUMIDITY=aht_humidity.relative_humidity + SETTING_HUMIDITY_COMP;
      TEMPERATURE_C=aht_temp.temperature + (SETTING_TEMP_COMP/float(10));
      TEMPERATURE_F=(TEMPERATURE_C * 9/5) + 32;
    } else if ( SETTING_TEMPHUM == 3 && timeinfo.tm_sec%10 == 0 ) { // READ EVERY 10 SECONDS
      sensors_event_t aht_humidity, aht_temp;
      aht.getEvent(&aht_humidity, &aht_temp);// populate temp and humidity objects with fresh data
      HUMIDITY=aht_humidity.relative_humidity + SETTING_HUMIDITY_COMP;
      TEMPERATURE_C=aht_temp.temperature + (SETTING_TEMP_COMP/float(10));
      TEMPERATURE_F=(TEMPERATURE_C * 9/5) + 32;
    }
  }
}

void display_date() {
  if ( SETTING_DATEDISP_SEC != 0 ) {
    if ( timeinfo.tm_sec%SETTING_DATEDISP_SEC < SETTING_DATEDISP_DUR ) {
      switch (SETTING_DATEFORMAT) {
         case 2:
          sec_ones = timeinfo.tm_mday % 10;
          sec_tens = timeinfo.tm_mday / 10;
          min_ones = (timeinfo.tm_mon+1) % 10;
          min_tens = (timeinfo.tm_mon+1) / 10;
          hour_ones = (timeinfo.tm_year-100) % 10;
          hour_tens = (timeinfo.tm_year-100) / 10;
          break;
         case 3:
          sec_ones = (timeinfo.tm_year-100) % 10;
          sec_tens = (timeinfo.tm_year-100) / 10;
          min_ones = (timeinfo.tm_mon+1) % 10;
          min_tens = (timeinfo.tm_mon+1) / 10;
          hour_ones = timeinfo.tm_mday % 10;
          hour_tens = timeinfo.tm_mday / 10;
          break;          
         default: // THIS IS CASE 1, our default case
          sec_ones = (timeinfo.tm_year-100) % 10;
          sec_tens = (timeinfo.tm_year-100) / 10;
          min_ones = timeinfo.tm_mday % 10;
          min_tens = timeinfo.tm_mday / 10;
          hour_ones = (timeinfo.tm_mon+1) % 10;
          hour_tens = (timeinfo.tm_mon+1) / 10;
          break;
      }
    }
  }
}

void GET_ESP_PARITIONS() {
  char PARTOUTPUT[500];
  //ESP_LOGI(TAG, "Start test");
  esp_partition_iterator_t iter = esp_partition_find(ESP_PARTITION_TYPE_APP, ESP_PARTITION_SUBTYPE_ANY, NULL);
  //ESP_LOGI(TAG, "Name, type, subtype, offset, length");
  snprintf(PARTOUTPUT, sizeof(PARTOUTPUT), "Name, type, subtype, offset, length");
  while (iter != nullptr)
  {
    const esp_partition_t *partition = esp_partition_get(iter);
    //ESP_LOGI(TAG, "%s, app, %d, 0x%x, 0x%x (%d)", partition->label, partition->subtype, partition->address, partition->size, partition->size);
    snprintf(PARTOUTPUT, sizeof(PARTOUTPUT), "%s<br>%s, app, %d, 0x%x, 0x%x (%d)", PARTOUTPUT, partition->label, partition->subtype, partition->address, partition->size, partition->size);
    iter = esp_partition_next(iter);
  }
  
  esp_partition_iterator_release(iter);
  iter = esp_partition_find(ESP_PARTITION_TYPE_DATA, ESP_PARTITION_SUBTYPE_ANY, NULL);
  while (iter != nullptr)
  {
    const esp_partition_t *partition = esp_partition_get(iter);
    //ESP_LOGI(TAG, "%s, data, %d, 0x%x, 0x%x (%d)", partition->label, partition->subtype, partition->address, partition->size, partition->size);
    snprintf(PARTOUTPUT, sizeof(PARTOUTPUT), "%s<br>%s, data, %d, 0x%x, 0x%x (%d)", PARTOUTPUT, partition->label, partition->subtype, partition->address, partition->size, partition->size);
    iter = esp_partition_next(iter);
  }    
  esp_partition_iterator_release(iter);
  
  Server.send(200, "text/html", PARTOUTPUT);
}

void setup() {
  esp_task_wdt_init(300, true);  // ENABLE WATCHDOG
  esp_task_wdt_add(NULL);       // ADD CURRENT THREAD
  
  pinMode(CLOCK_PIN, OUTPUT);
  pinMode(DATA_PIN, OUTPUT);

  #ifdef IV_9_6D_RR
    pinMode(RCK_PIN, OUTPUT);
  #endif

  // SET UP PWM ON OUTPUT ENABLE PIN FOR DIMMING
  //ledcSetup(0, 400, 8);   //  ledcSetup(uint8_t channel, double freq, uint8_t resolution_bits);
  ledcSetup(0, 1600, 8);   //  ledcSetup(uint8_t channel, double freq, uint8_t resolution_bits);
  ledcAttachPin(OE_PIN, 0);
  ledcWrite(0, 0);    // INVERTED (0, 0)   // REGULAR (0, 255)

  BRIGHTNESS=255;   // ADDED in 0.91v BETA4, without this when the light meter was enabled, but no light meter was installed, the brightness was never set and display would be blank

  // CLEAR OUTPUT BITS OF SHIFT REGISTER
  int var = 0;
  while (var <= 65) {
    WRITE_ONELOWBIT();    // NO NEED TO WORRY ABOUT INVERTING HERE, IT'S DONE IN THE FUNCTION BOTH LOWBIT AND HIGHBIT
    var++;
  }

  // SETTINGS THAT GET STORED IN FLASH
  preferences.begin("clock-settings", false);

  TOTAL_RUNTIME = preferences.getFloat("TOTAL_RUNTIME", 0);   // Get TOTAL_RUNTIME before processing possible reset of variables below

  pinMode(EEPROM_RESET_PIN, INPUT_PULLUP);           // READ PIN (set as a define, check beginning of file) and if low (ground) reset permanent (eeprom/flash) variables to their default values, except TOTAL_RUNTIME
  if ( digitalRead(EEPROM_RESET_PIN) == 0 ) {
    preferences.clear();
    WRITE_DISPLAY(0,0,0,0,0,0,0,0,0,0,0,0); delay(500); WRITE_DISPLAY(DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,0,0,0,0,0,0); delay(500);
    WRITE_DISPLAY(0,0,0,0,0,0,0,0,0,0,0,0); delay(500); WRITE_DISPLAY(DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,0,0,0,0,0,0); delay(500);
    WRITE_DISPLAY(0,0,0,0,0,0,0,0,0,0,0,0); delay(500); WRITE_DISPLAY(DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,DIGIT_BLANK,0,0,0,0,0,0); delay(500);
  }
    
  preferences.putFloat("TOTAL_RUNTIME", TOTAL_RUNTIME);   // Put TOTAL_RUNTIME back into permanent storage, as we don't want to reset this value

  #ifdef IV_9_6D_RR
    SETTING_BRIGHTNESS = preferences.getInt("BRIGHTNESS", 165);
  #endif
  #ifdef Z57XM6DV2
    SETTING_BRIGHTNESS = preferences.getInt("BRIGHTNESS", 255);
  #endif
  SETTING_LUXMETER = preferences.getInt("LUXMETER", 1);
  //SETTING_DISPLAYTEMP = preferences.getInt("DISPLAYTEMP", 0);
  SETTING_TEMPDISP_SEC = preferences.getInt("TEMPDISP_SEC", 0);
  SETTING_TEMPDISP_DUR = preferences.getInt("TEMPDISP_DUR", 2);
  
  SETTING_TEMP_IN_C = preferences.getInt("TEMP_IN_C", 0);
  SETTING_TEMP_COMP = preferences.getInt("TEMP_COMP", 0);
  SETTING_HUMIDITY_COMP = preferences.getInt("HUMIDITY_COMP", 0);
  SETTING_12H = preferences.getInt("12H", 0);
  SETTING_LZBLANK = preferences.getInt("LZBLANK", 0);

  SETTING_DATEFORMAT = preferences.getInt("DATEFORMAT", 1);
  SETTING_DATEDISP_SEC = preferences.getInt("DATEDISP_SEC", 0);
  SETTING_DATEDISP_DUR = preferences.getInt("DATEDISP_DUR", 2);
    
  SETTING_CATHODEPP = preferences.getInt("CATHODEPP", 1);
  SETTING_CATHODEPPSTART = preferences.getInt("CATHODEPPSTART", 3);
  SETTING_CATHODEPPDUR = preferences.getInt("CATHODEPPDUR", 1);
  SETTING_CATHODEPPREV = preferences.getInt("CATHODEPPREV", 10);
  SETTING_CATHODEPPREVD = preferences.getInt("CATHODEPPREVD", 4);
  
  SETTING_TIMEZONE = preferences.getString("TIMEZONE", "PST8PDT,M3.2.0,M11.1.0");

  SETTING_TUBEONBOTTOM = preferences.getInt("TUBEONBOTTOM", 1);

  SETTING_COLON = preferences.getInt("COLON", 1);

  SETTING_BMP = preferences.getInt("BMP", 1);

  SETTING_TEMPHUM = preferences.getInt("TEMPHUM", 1);

  SETTING_ALTDISPLAY = preferences.getInt("ALTDISPLAY", 1);
  
  preferences.end();

  // ORIGINAL SEC_ONE_MAP[], SEC_TEN_MAP[], and so on... HV5x222 to Nixie Tube digit DEFINITIONS WHERE HERE, NOW IN z57xm6d-definitions.h 

  demo_digits();

  //char tmp_hostname[30];    // MOVED TO BE DEFINED AS GLOBAL 
  uint8_t mac[6];
  WiFi.macAddress(mac);
  #ifdef IV_9_6D_RR
    sprintf(tmp_hostname, "iv9clock-%02x%02x", mac[4], mac[5]);   // BETA4 REMOVED mac[0] - mac[4] to shorten hostname
  #endif
  #ifdef Z57XM6DV2
    sprintf(tmp_hostname, "Z57clock-%02x%02x", mac[4], mac[5]);   // BETA4 REMOVED mac[0] - mac[4] to shorten hostname
  #endif

  WiFi.mode(WIFI_MODE_NULL);      // HAD TO ADD THIS FOR WiFi.setHostname to work! 12-26-2022  See bug... https://github.com/espressif/arduino-esp32/issues/6700
  WiFi.setHostname(tmp_hostname); // THIS IS THE HOSTNAME REPORTED TO DHCP (usually the one that shows up in your router 'clients' list
  
  // Initialize Serial Monitor
  Serial.begin(115200);
  serial_debugln();
  Server.on("/", rootPage);
  Server.on("/partitions.html", GET_ESP_PARITIONS);
  //config.hostName = "z57clock";
  config.hostName = "tmp_hostname";

  #ifdef IV_9_6D_RR
    config.apid = "iv9clock";
  #else
    config.apid = "z57clock";
  #endif
  config.psk = "neonixie.com";

  config.auth = AC_AUTH_DIGEST;
  config.authScope = AC_AUTHSCOPE_PARTIAL;    // This configures authentication only on the web based firmware update html page (autoconnect.h)
  #ifdef IV_9_6D_RR
    config.username = "iv9clock";
    config.password = "neonixie.com";
  #endif
  #ifdef Z57XM6DV2
    config.username = "z57clock";
    config.password = "neonixie.com";
  #endif
  config.ota = AC_OTA_BUILTIN;  // Turns on web based firmware update (built into autoconnect.h)

  config.menuItems = config.menuItems | AC_MENUITEM_DELETESSID; // THIS ADDS THE OPTION TO DELETE AN SSID 

  config.autoReconnect = true;    // Without this option, when the wifi router was shut off and turned back on, the clock would not attempt a connection again to the wifi, it would just go into soft AP mode. Adding this fixed it and it now attempts to reconnect when the wifi router is back online. This was clued in from https://github.com/Hieromon/AutoConnect/issues/80
  Portal.config(config);
  if (Portal.begin()) {
    serial_debugln("HTTP server:" + WiFi.localIP().toString());    
  }

  // Initialize a NTPClient to get time   // THIS IS THE PRIOR OLDER STUFF, NO LONGER USED REALLY
  timeClient.begin();
  timeClient.setTimeOffset(-28800);   // Set offset time in seconds to adjust for your timezone, for example: GMT +1 = 3600     GMT +8 = 28800    GMT -1 = -3600    GMT 0 = 0

// ADD PAGE TEST
  delay(1000);
  page_add.load(PAGE_ADD);
  //page_results.load(PAGE_RESULTS);
  //Portal.join({ page_add, page_results });
  Portal.join({ page_add });
  Portal.on("/clockconfig", onclockconfig);
  //Portal.on("/results", onResults);
  Portal.begin();
// END ADD PAGE TEST


  // DISPLAY LOCAL IP ADDRESS
  WRITE_DISPLAY(WiFi.localIP()[0]/100, WiFi.localIP()[0]%100/10, WiFi.localIP()[0]%10, WiFi.localIP()[1]/100, WiFi.localIP()[1]%100/10, WiFi.localIP()[1]%10, 0, 0, 1, 0, 0, 0);
  delay(4000);
  WRITE_DISPLAY(WiFi.localIP()[2]/100, WiFi.localIP()[2]%100/10, WiFi.localIP()[2]%10, WiFi.localIP()[3]/100, WiFi.localIP()[3]%100/10, WiFi.localIP()[3]%10, 0, 0, 1, 0, 0, 0);
  delay(4000);    

  initTime(SETTING_TIMEZONE);   // PST8PDT,M3.2.0,M11.1.0   // THIS IS THE ACTUAL ONE THAT WORKS WITH TIMEZONES NOW 06/04/2022
  printLocalTime();

  // OTA STUFF START

  // Port defaults to 3232
  // ArduinoOTA.setPort(3232);

  // Hostname defaults to esp3232-[MAC]
  // ArduinoOTA.setHostname("myesp32");

  //char tmp_hostname[30];    // MOVED ABOVE
  //uint8_t mac[6];
  //WiFi.macAddress(mac);
  //sprintf(tmp_hostname, "Z57clock-%02x%02x", mac[4], mac[5]);   // BETA4 REMOVED mac[0] - mac[4] to shorten hostname
  
  ArduinoOTA.setHostname(tmp_hostname);

  // No authentication by default
  // ArduinoOTA.setPassword("admin");
  ArduinoOTA.setPassword("neonixie.com");

  // Password can be set with it's md5 value as well
  // MD5(admin) = 21232f297a57a5a743894a0e4a801fc3
  // ArduinoOTA.setPasswordHash("21232f297a57a5a743894a0e4a801fc3");

        ArduinoOTA
    .onStart([]() {
      String type;
      if (ArduinoOTA.getCommand() == U_FLASH)
        type = "sketch";
      else // U_SPIFFS
        type = "filesystem";

      // NOTE: if updating SPIFFS this would be the place to unmount SPIFFS using SPIFFS.end()
      serial_debugln("Start updating " + type);
    })
    .onEnd([]() {
      serial_debugln("\nEnd");
    })
    .onProgress([](unsigned int progress, unsigned int total) {
      serial_debug_printf("Progress: %u%%\r", (progress / (total / 100)));
    })
    .onError([](ota_error_t error) {
      serial_debug_printf("Error[%u]: ", error);
      if (error == OTA_AUTH_ERROR) serial_debugln("Auth Failed");
      else if (error == OTA_BEGIN_ERROR) serial_debugln("Begin Failed");
      else if (error == OTA_CONNECT_ERROR) serial_debugln("Connect Failed");
      else if (error == OTA_RECEIVE_ERROR) serial_debugln("Receive Failed");
      else if (error == OTA_END_ERROR) serial_debugln("End Failed");
    });

  ArduinoOTA.begin();
  // OTA STUFF END

  Wire.begin();
  lightMeter.begin();

  // START AHT20 TEMP HUMIDITY SENSOR
  aht.begin();

  #ifdef BMP_ENABLE
  // BMP280 sensor
  bmp.begin(0x77);
  /* Default settings from datasheet. */
  bmp.setSampling(Adafruit_BMP280::MODE_NORMAL,     /* Operating Mode. */
    Adafruit_BMP280::SAMPLING_X2,     /* Temp. oversampling */
    Adafruit_BMP280::SAMPLING_X16,    /* Pressure oversampling */
    Adafruit_BMP280::FILTER_X16,      /* Filtering. */
    Adafruit_BMP280::STANDBY_MS_500); /* Standby time. */
  #endif
}

// WIFI RECONNECT TEST // ADDED 01/10/2022, to see if it fixes reconnect problem. // 07/11/2022 SEEMS TO RECONNECT WHEN NEEDED
unsigned long previousMillis = 0;
unsigned long interval = 30000;

unsigned long NTP_previousMillis = 9999999;   // MUST BE LARGER THEN interval
unsigned long NTP_interval = 300000;   // INTERVAL IN MILLISECONDS TO RUN THE timeClient.update() FUNCTION.  60000 = 1 minute  300000 = 5 minutes

void loop() {
  
  // NEW STUFF WITH TIMEZONE
  //printLocalTime();
  //struct tm timeinfo;   // MOVED TO GLOBAL
  if(!getLocalTime(&timeinfo)){
    serial_debugln("Failed to obtain time 1");
    return;
  }
  
  ArduinoOTA.handle();    // OTA
  
  Portal.handleClient();
  
  unsigned long NTP_currentMillis = millis();
  if ( NTP_currentMillis - NTP_previousMillis >=NTP_interval ) {
    //Serial.print("timeClient.update start");
    serial_debug("timeClient.update start");
    timeClient.update();      // NOTE, POSSIBLY DUE TO DNS DELAYS OR ?? THIS FUNCTION WILL SOMETIMES DELAY BY SEVERAL SECONDS. SET TO RUN ONLY PERIODICALLY SO AS NOT TO DISTRUPT THE CLOCK DISPLAY
    //Serial.print("timeClient.update end");
    serial_debug("timeClient.update end");
    NTP_previousMillis = NTP_currentMillis;
  }

  sec_ones = timeinfo.tm_sec % 10;
  sec_tens = timeinfo.tm_sec / 10;
  min_ones = timeinfo.tm_min % 10;
  min_tens = timeinfo.tm_min / 10;
  if ( SETTING_12H == 1 ) {
    if ( timeinfo.tm_hour == 0 ) {  // MIDNIGHT, AM
      hour_ones=2;
      hour_tens=1;
    } else if ( timeinfo.tm_hour >= 1 && timeinfo.tm_hour <= 11 ) { // 1 to 11 AM
      hour_ones = timeinfo.tm_hour % 10;
      hour_tens = timeinfo.tm_hour / 10;
    } else if ( timeinfo.tm_hour == 12 ) {  // NOON, PM
      hour_ones=2;
      hour_tens=1;
    } else {    // THIS IS ANYTHING AFTER 1300
      hour_ones = ((timeinfo.tm_hour)-12) % 10;
      hour_tens = ((timeinfo.tm_hour)-12) / 10; 
    }
  } else {
    hour_ones = timeinfo.tm_hour % 10;
    hour_tens = timeinfo.tm_hour / 10;
  }
  
  if ( SETTING_LZBLANK == 1 ) {
    if ( hour_tens == 0 ) {
      hour_tens=DIGIT_BLANK;
    }
  }

  if ( sec_ones % 2 != 0 && SETTING_COLON == 2 ) {    // DECIMAL POINT ON 
    hour_tens_dp = 0; hour_ones_dp = 1; min_tens_dp = 0; min_ones_dp = 1; sec_tens_dp = 0; sec_ones_dp = 0;
  } else {
    hour_tens_dp = 0; hour_ones_dp = 0; min_tens_dp = 0; min_ones_dp = 0; sec_tens_dp = 0; sec_ones_dp = 0;
  }

// WIFI RECONNECT TEST
// ADDED 01/10/2022, to see if it fixes reconnect problem. // 05/29/2022 SEEMS TO WORK WELL
unsigned long currentMillis = millis();
// if WiFi is down, try reconnecting
if ((WiFi.status() != WL_CONNECTED) && (currentMillis - previousMillis >=interval)) {
  //Serial.print(millis());
  serial_debug(millis());
  serial_debugln("Reconnecting to WiFi...");
  WiFi.disconnect();
  WiFi.reconnect();
  previousMillis = currentMillis;
}

// START ONCE PER SECOND LOOP
if ( previous_tm_sec != timeinfo.tm_sec ) { 
  previous_tm_sec=timeinfo.tm_sec;

  esp_task_wdt_reset();   // Watchdog timer reset

  TOTAL_RUNTIME_SECONDS_COUNTER++;
  if ( TOTAL_RUNTIME_SECONDS_COUNTER >= 864 ) {  // 1/100th of a day has passed, increment UPTIME and TOTAL_RUNTIME, save to permanent memory, DON'T UPDATE MEMORY TOO OFTEN due to flash wear!
    TOTAL_RUNTIME_SECONDS_COUNTER=0;
    UPTIME=UPTIME + 0.01;
    TOTAL_RUNTIME=TOTAL_RUNTIME + 0.01;
    preferences.begin("clock-settings", false);
    preferences.putFloat("TOTAL_RUNTIME", TOTAL_RUNTIME);
    preferences.end();
  }

  if ( SETTING_LUXMETER == 2 ) {  
    // READ BL1750 LIGHT METER
    lux = lightMeter.readLightLevel();
    //Serial.print("Light: "); Serial.print(lux); Serial.print(" lx"); Serial.print(" ");
    serial_debug("Light: "); serial_debug(lux); serial_debug(" lx"); serial_debug(" ");

    #ifdef IV_9_6D_RR
    // SET PWM ON OUTPUT PIN FOR DIMMING
    if ( lux < 0 ) {
      // do nothing! something is wrong! Sometimes we get a -1 error reading from the sensor
    } else if ( lux > 50 ) {
      //ledcWrite(0, 0);      // FULL BRIGHTNESS  // (0, 0) for inverted output, (0, 255) for regular
      BRIGHTNESS=SETTING_BRIGHTNESS;
    } else if ( lux > 5 ) {
      //ledcWrite(0, 128);    // HALF BRIGHT      // (0, 150 for inverted output, (0, 100) for regular
      BRIGHTNESS=((SETTING_BRIGHTNESS-100)/2) + 100; // SET MEDIUM BRIGHTNESS MIDWAY BETWEEN SETTING_BRIGHTNESS AND DIM BRIGHT
    } else {
      //ledcWrite(0, 205);    // DIM              // (0, 200) for inverted output, (0, 50) for regular
      BRIGHTNESS=100;
    }
    #endif

    #ifdef Z57XM6DV2
    // SET PWM ON OUTPUT PIN FOR DIMMING
    if ( lux < 0 ) {
      // do nothing! something is wrong! Sometimes we get a -1 error reading from the sensor
    } else if ( lux > 50 ) {
      //ledcWrite(0, 0);      // FULL BRIGHTNESS  // (0, 0) for inverted output, (0, 255) for regular
      BRIGHTNESS=255;
    } else if ( lux > 5 ) {
      //ledcWrite(0, 128);    // HALF BRIGHT      // (0, 150 for inverted output, (0, 100) for regular
      BRIGHTNESS=128;
    } else {
      //ledcWrite(0, 205);    // DIM              // (0, 200) for inverted output, (0, 50) for regular
      BRIGHTNESS=50;
    }
    #endif
    
  } else if ( SETTING_LUXMETER == 3 ) {   // JUST READ THE LUXMETER, DON'T ADJUST BRIGHTNESS, MAINLY TO DISPLAY THE LUX ON THE WEB INTERFACE
    // READ BL1750 LIGHT METER
    lux = lightMeter.readLightLevel();
    //Serial.print("Light: "); Serial.print(lux); Serial.print(" lx"); Serial.print(" ");
    serial_debug("Light: "); serial_debug(lux); serial_debug(" lx"); serial_debug(" ");
    BRIGHTNESS=SETTING_BRIGHTNESS;
  } else {
    //ledcWrite(0, abs(SETTING_BRIGHTNESS-255));  // INVERTED BRIGHTNESS
    //BRIGHTNESS=abs(SETTING_BRIGHTNESS-255);
    BRIGHTNESS=SETTING_BRIGHTNESS;
  }

  #ifdef BMP_ENABLE
    read_bmp(); // READ BMP280 sensor, we currently just display it on the main webpage
  #endif
  display_temperature();
  display_date();

  //strftime(TIME_STRING, sizeof(TIME_STRING), "%A, %B %d %Y %H:%M:%S %F (%Z) weekday(sun): %u weeknumber(sun): %U weekday(mon): %w weeknumber(mon): %W TimeZoneOffset: %z ", &timeinfo);
  strftime(TIME_STRING, sizeof(TIME_STRING), "%A, %B %d %Y %H:%M:%S %F %Z %z", &timeinfo);
  //Serial.print(&timeinfo, "%A, %B %d %Y %H:%M:%S zone %Z %z ");
  serial_debug_x2(&timeinfo, "%A, %B %d %Y %H:%M:%S zone %Z %z ");   // ERROR, 2 arguments
  //Serial.println("");   // PUT A FINAL RETURN CHARACTER OUT THE SERIAL PORT
  serial_debugln("");   // PUT A FINAL RETURN CHARACTER OUT THE SERIAL PORT

  #ifdef Z57XM6DV2
  if ( SETTING_CATHODEPP != 1 ) {
    cathodepp();
  }
  #endif

  ledcWrite(0, abs(BRIGHTNESS-255));
  if ( SETTING_ALTDISPLAY == 2 ) {  // ALTIMETER
    sec_ones = int(bmp_FEET*10)%10; sec_tens = int(bmp_FEET)%10; min_ones = int(bmp_FEET)%100/10; min_tens = int(bmp_FEET)%1000/100; hour_ones = int(bmp_FEET)%10000/1000; hour_tens = int(bmp_FEET)/10000;
    hour_tens_dp = 0; hour_ones_dp = 0; min_tens_dp = 0; min_ones_dp = 0; sec_tens_dp = 1; sec_ones_dp = 0;
  } else if ( SETTING_ALTDISPLAY == 3 ) { // RSSI
    hour_tens = DIGIT_BLANK; hour_ones = DIGIT_BLANK; min_tens = DIGIT_BLANK; min_ones = DIGIT_NEGATIVE; sec_tens = (abs(WiFi.RSSI()))/10; sec_ones = int((abs(WiFi.RSSI())))%10;
    hour_tens_dp = 0; hour_ones_dp = 0; min_tens_dp = 0; min_ones_dp = 0; sec_tens_dp = 0; sec_ones_dp = 0;
  } else if ( SETTING_ALTDISPLAY == 4 ) { // LUX
    //sec_ones = int(lux)%10; sec_tens = int(lux)%100/10; min_ones = int(lux)%1000/100; min_tens = int(lux)%10000/1000; hour_ones = int(lux)/10000; hour_tens = 10;   // DISPLAY LUX LEVELS
    hour_tens = DIGIT_BLANK; hour_ones = int(lux)/10000; min_tens = int(lux)%10000/1000; min_ones = int(lux)%1000/100; sec_tens = int(lux)%100/10; sec_ones = int(lux)%10;
    hour_tens_dp = 0; hour_ones_dp = 0; min_tens_dp = 0; min_ones_dp = 0; sec_tens_dp = 0; sec_ones_dp = 0;
  } else if ( SETTING_ALTDISPLAY == 5 ) { // BAROMETER
    snprintf(bmp_HG_STRING_BUFFER, sizeof(bmp_HG_STRING_BUFFER), "%.2f", bmp_HG);   // FOR PROPER ROUNDING OF bmp_HG
    #ifdef IV_9_6D_RR
      hour_tens = int(bmp_HG)%100/10; hour_ones = int(bmp_HG)%10; min_tens = int( atof(bmp_HG_STRING_BUFFER) *10)%10; min_ones = int( atof(bmp_HG_STRING_BUFFER) *100)%10; sec_tens = int('H')-55; sec_ones = int('G')-55;
    #else    // Z57 clock cannot display letters
      hour_tens = int(bmp_HG)%100/10; hour_ones = int(bmp_HG)%10; min_tens = int( atof(bmp_HG_STRING_BUFFER) *10)%10; min_ones = int( atof(bmp_HG_STRING_BUFFER) *100)%10; sec_tens = DIGIT_BLANK; sec_ones = DIGIT_BLANK;
    #endif
    hour_tens_dp = 0; hour_ones_dp = 1; min_tens_dp = 0; min_ones_dp = 0; sec_tens_dp = 1; sec_ones_dp = 1;
  }

  WRITE_DISPLAY(hour_tens, hour_ones, min_tens, min_ones, sec_tens, sec_ones, hour_tens_dp, hour_ones_dp, min_tens_dp, min_ones_dp, sec_tens_dp, sec_ones_dp);
} 
// END ONCE PER SECOND LOOP
  
}
